package ninja.thefirearchmage.games.fourtykcalculator;

import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple3;

import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * Models what modifiers a unit that is performing the attacking provides to its weapons.
 * For example, a unit of eradicators has reroll on hits, wounds and damage against monsters and vehicles.
 */
@NoArgsConstructor @Getter @Setter
public class UnitAttackModifiers {
    private Tuple3<Boolean, WeaponRangeType, Integer> sustainedHitsModifier;
    private Tuple2<Boolean, WeaponRangeType> lethalHitsModifier;
    private int criticalHitValue = 6;
    private boolean lanceModifier;
    // These rerolls do not contain actual values, they must be generated from the method getRerollFor
    private List<Reroll> rerollsModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> apModifier;
    private Tuple2<Boolean, WeaponRangeType> woundBonusModifier;
    private Tuple2<Boolean, WeaponRangeType> woundMalusModifier;
    private Tuple2<Boolean, WeaponRangeType> hitBonusModifier;
    private Tuple2<Boolean, WeaponRangeType> hitMalusModifier;
    private Tuple2<Boolean, WeaponRangeType> devastatingWoundsModifier;
    private Tuple2<Boolean, WeaponRangeType> hazardousModifier;
    private Tuple2<Boolean, WeaponRangeType> heavyModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> strengthModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> attacksBonusModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> attacksMalusModifier;
    private Tuple3<Boolean, WeaponRangeType, Integer> rapidFireModifier;
    private boolean unmodifiableHitRoll;
    private boolean fishForCriticalHits;
    private boolean fishForCriticalWounds;

    private boolean unitHasRemainedStationaryModifier;
    private boolean unitCharged;
    private boolean targetInRapidFireRange;
    private boolean isPsychic;
    private boolean canRerollHazardous;

    public boolean hasRerollFor(RerollStageType stageType, RerollType type,
                                Set<UnitTypeKeyword> targetsType, WeaponRangeType weaponRangeType) {
        return rerollsModifier.stream()
                .filter( r -> setContainsAnyOf(r.getValidTargetTypes(), targetsType, UnitTypeKeyword.ALL))
                .filter( r -> setContainsAnyOf(r.getValidWeaponRanges(), Set.of(weaponRangeType), WeaponRangeType.ALL))
                .filter(r -> r.getRerollStageType() == stageType)
                .anyMatch(r -> r.getType() == type);
    }

    public Optional<Reroll> getRerollFor(RerollStageType stageType, RerollType type,
                                         Set<UnitTypeKeyword> targetsType, WeaponRangeType weaponRangeType, int rollGoal) {
        if(hasRerollFor(stageType, type, targetsType, weaponRangeType)){
            return Optional.of(switch(type) {
                case ON_ONES -> Reroll.rerollOnOnes(stageType, targetsType, Set.of(weaponRangeType));
                case ON_FAILURE -> Reroll.rerollFailures(stageType, rollGoal, targetsType, Set.of(weaponRangeType));
                case ANY -> Reroll.rerollOnNot(stageType, IntStream.range(rollGoal,7).boxed().collect(Collectors.toSet()), targetsType, Set.of(weaponRangeType));
            });
        }

        return Optional.empty();
    }

    private static <T> boolean setContainsAnyOf(Set<T> oneSet, Set<T> otherSet, T otherValue) {
        return oneSet.stream().anyMatch(i -> i.equals(otherValue) || otherSet.contains(i));
    }

    public boolean lethalHitsAppliesTo(Weapon weapon) {
        return lethalHitsModifier._1 && lethalHitsModifier._2.appliesTo(weapon);
    }
}
