package ninja.thefirearchmage.games.fourtykcalculator;

import lombok.Getter;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple;
import ninja.thefirearchmage.games.fourtykcalculator.utils.datastructures.Tuple2;

import java.util.*;
import java.util.stream.Collectors;

public class Weapon {
    @Getter
    private String name;
    @Getter
    private int strength;
    @Getter
    private int ap;
    @Getter
    private int attacks;
    @Getter
    private WeaponDamage damage;
    @Getter
    private WeaponRangeType rangeType;
    @Getter
    private int hitValue;
    private boolean hasLethalHits;
    @Getter
    private Optional<Integer> sustainedHits;
    private boolean hasDevastatingWounds;
    private boolean hasHazardous;
    private List<AntiUnitEffect> antiUnitEffects;
    private boolean hasHeavy;
    private boolean hasBlast;
    private boolean hasLance;
    private boolean ignoresCover;
    private boolean hasIndirectFire;
    @Getter
    private Optional<Integer> melta;
    @Getter
    private Optional<Integer> rapidFire;
    private boolean isPsychic;

    public Weapon(String name, int strength, int ap, int attacks, int hitValue, WeaponDamage damage,
                  WeaponRangeType weaponRangeType) {
        this.name = name;
        this.strength = strength;
        this.ap = ap;
        this.attacks = attacks;
        this.hitValue = hitValue;
        this.damage = damage;
        this.rangeType = weaponRangeType;

        this.antiUnitEffects = new ArrayList<>();
        this.melta = Optional.empty();
        this.rapidFire = Optional.empty();
        this.sustainedHits = Optional.empty();

        this.hasDevastatingWounds = false;
        this.hasHazardous = false;
        this.hasHeavy = false;
        this.hasBlast = false;
        this.hasLance = false;
        this.ignoresCover = false;
        this.hasIndirectFire = false;
        this.hasLethalHits = false;
    }

    /**
     * Given a defensive profile, attack unit modifiers and the number of weapons to shoot, resolves the attack
     * portion statistics of the attack scenario. Returns the stats about the attack itself, before the wounds
     * are assigned to the target.
     */
    public WeaponAttackResolution generateAttackResolution(DefensiveProfile target, UnitAttackModifiers unitModifiers,
                                                           int weaponAmount) {
        var weaponAttackResolution = new WeaponAttackResolution();

        var passingHits = calculatePassingHits(target, unitModifiers, weaponAmount, weaponAttackResolution);
        var passingWounds = calculatePassingWounds(target, unitModifiers, weaponAmount, passingHits, weaponAttackResolution);
        var normalWounds = passingWounds._1;
        var mortalWounds = passingWounds._2;

        return weaponAttackResolution;
    }


    private double calculatePassingHits(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers,
                                        int amountOfWeapons, WeaponAttackResolution weaponAttackResolution) {
        int hitRollGoal = calculateHitRollGoal(target, unitAttackModifiers);
        double hitSuccessProbability = (double) calculateProbabilityNumeratorFrom(hitRollGoal) / 6;
        double hitFailureProbability = (double) (6 - calculateProbabilityNumeratorFrom(hitRollGoal)) / 6;
        int weaponAttacks = calculateWeaponAttacks(amountOfWeapons, unitAttackModifiers, target, weaponAttackResolution);

        // Base passing hits without rerolls
        double passingHits = weaponAttacks * hitSuccessProbability;
        weaponAttackResolution.setHitsWithoutRerolls(passingHits);

        // Add hits from rerolls
        var hitRerolls = prepareRerolls(RerollStageType.HIT, hitRollGoal, unitAttackModifiers, target);
        if(hitRerolls.isPresent()) {
            var rerolls = hitRerolls.get();
            var successfulHitFromRerolls = weaponAttacks * hitFailureProbability * rerolls.getSuccessProbability();
            weaponAttackResolution.setHitsFromRerolls(successfulHitFromRerolls);
            passingHits += successfulHitFromRerolls;
        }

        // calculate sustained hits
        var sustainedHits = calculateSustainedHits(weaponAttacks, unitAttackModifiers, target);
        weaponAttackResolution.setHitsFromSustainedHits(sustainedHits);
        passingHits += sustainedHits;

        weaponAttackResolution.setHits(passingHits);

        return passingHits;
    }

    private double calculateSustainedHits(int weaponAttacks, UnitAttackModifiers unitAttackModifiers,
                                          DefensiveProfile target) {
        var sustainedHitsAmounts = 0d;

        var hasSustainedHits = this.sustainedHits.isPresent() ||
                (unitAttackModifiers.getSustainedHitsModifier()._1 && unitAttackModifiers.getSustainedHitsModifier()._2.appliesTo(this) );
        if(hasSustainedHits) {
            var weaponSustainedHits = this.sustainedHits.orElse(0);
            var unitModifiersSustainedHits = unitAttackModifiers.getSustainedHitsModifier()._1 ?
                    unitAttackModifiers.getSustainedHitsModifier()._3 : 0;
            var sustainedHitsValue = Math.max(weaponSustainedHits, unitModifiersSustainedHits);
            var criticalHitValue = unitAttackModifiers.getCriticalHitValue();

            double nonSustainedProbability = (double) (6 - (7 - criticalHitValue)) / 6;
            double sustainedProbability = (double) (7 - criticalHitValue) / 6;
            var hasHitRerollAny = unitAttackModifiers.hasRerollFor(RerollStageType.HIT, RerollType.ANY,
                    target.getUnitTypeKeywords(), this.rangeType);
            if(unitAttackModifiers.isFishForCriticalHits() && hasHitRerollAny) {
                sustainedProbability += nonSustainedProbability*sustainedProbability;
            }
            sustainedHitsAmounts = weaponAttacks * sustainedProbability * sustainedHitsValue;
        }

        return sustainedHitsAmounts;
    }

    /**
     * Returns a tuple of normal passing wounds and mortal wounds so that the caller can handle them differently.
     */
    private Tuple2<Double, Double> calculatePassingWounds(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers,
                                                          int amountOfWeapons, double passingHits,
                                                          WeaponAttackResolution weaponAttackResolution) {
        double lethalHitsCount = calculateLethalHits(amountOfWeapons, unitAttackModifiers, target, weaponAttackResolution);
        var normalHits = passingHits - lethalHitsCount;

        int woundRollGoal = calculateWoundRollGoal(target, unitAttackModifiers);
        int woundSuccessProbabilityNumerator = calculateProbabilityNumeratorFrom(woundRollGoal);

        var passingNormalWounds = normalHits * woundSuccessProbabilityNumerator / 6;
        weaponAttackResolution.setNormalWoundsWithoutRerolls(passingNormalWounds + lethalHitsCount);

        var woundsRerolls = prepareRerolls(RerollStageType.WOUND, woundRollGoal, unitAttackModifiers, target);
        if(woundsRerolls.isPresent()) {
            var rerolls = woundsRerolls.get();
            var failedWounds = normalHits * (6 - woundSuccessProbabilityNumerator) / 6;
            var extraWoundsFromRerolls = failedWounds * rerolls.getSuccessProbability() / 6;
            weaponAttackResolution.setNormalWoundsFromRerolls(extraWoundsFromRerolls);
            passingNormalWounds += extraWoundsFromRerolls;
        }

        // Mortal wounds
        var passingMortalWounds = 0d;
        if(weapon.hasDevastatingWounds()) {
            var criticalWoundRollGoal = weapon.getBestAntiFor(defensiveProfile)
                    .orElse(criticalWoundValue);
            var criticalProbability = calculateProbabilityNumeratorFrom(criticalWoundRollGoal);
            passingMortalWounds += (normalHits*criticalProbability/PROBABILITY_DENOMINATOR_D6);
            if(fishCriticalWounds && hasRerollWoundsAny) {
                passingMortalWounds += (normalHits * (1-criticalProbability)/PROBABILITY_DENOMINATOR_D6)*criticalProbability/PROBABILITY_DENOMINATOR_D6;
            }

            results.setWeaponStat(weapon, WeaponStat.MORTAL_WOUNDS_COUNT, passingMortalWounds);
            // passing wounds was calculated from the whole, but we must substract the critical wounds from the whole
            // to be able to handle them separately, as mortal wounds replace normal wounds.
            passingNormalWounds -= passingMortalWounds;
        }

        passingNormalWounds += lethalHitsCount;
        results.setWeaponStat(weapon, WeaponStat.WOUNDS_TOTAL_COUNT, passingNormalWounds+passingMortalWounds);

        return Tuple.of(passingNormalWounds, passingMortalWounds);
    }

    private double calculateLethalHits(int amountOfWeapons, UnitAttackModifiers unitAttackModifiers,
                                       DefensiveProfile target, WeaponAttackResolution weaponAttackResolution) {
        var lethalHitsCount = 0d;
        var hasLethalHits = this.hasLethalHits || unitAttackModifiers.lethalHitsAppliesTo(this);
        if(hasLethalHits) {
            double weaponAttacks = calculateWeaponAttacks(amountOfWeapons, unitAttackModifiers, target, weaponAttackResolution);
            var lethalHitsProbability = calculateProbabilityNumeratorFrom(unitAttackModifiers.getCriticalHitValue()) / 6;

            lethalHitsCount = weaponAttacks * lethalHitsProbability;
            // If we're fishing for hit criticals, the lethal hits will be higher
            if(unitAttackModifiers.isFishForCriticalHits()) {
                var lethalHitsFailureProbability = (6 - (lethalHitsProbability * 6)) / 6;
                var extraLethalHitsFromFishingCriticals = weaponAttacks * lethalHitsFailureProbability * lethalHitsProbability;
                weaponAttackResolution.setLethalHitsFromFishingCriticals(extraLethalHitsFromFishingCriticals);
                lethalHitsCount += extraLethalHitsFromFishingCriticals;
            }
            weaponAttackResolution.setNormalWoundsFromLethalHits(lethalHitsCount);
        }

        return lethalHitsCount;
    }

    private Optional<Reroll> prepareRerolls(RerollStageType stageType, int rollGoal,
                                            UnitAttackModifiers unitAttackModifiers, DefensiveProfile target) {
        var hasAnyReroll = unitAttackModifiers.hasRerollFor(stageType, RerollType.ANY, target.getUnitTypeKeywords(), this.rangeType);
        var hasOnesReroll = unitAttackModifiers.hasRerollFor(stageType, RerollType.ON_ONES, target.getUnitTypeKeywords(), this.rangeType);
        var hasFailedReroll = unitAttackModifiers.hasRerollFor(stageType, RerollType.ON_FAILURE, target.getUnitTypeKeywords(), this.rangeType);

        if (hasAnyReroll || hasFailedReroll) {
            return hasAnyReroll ?
                    unitAttackModifiers.getRerollFor(stageType, RerollType.ANY, target.getUnitTypeKeywords(), this.rangeType, rollGoal)
                    : unitAttackModifiers.getRerollFor(stageType, RerollType.ON_FAILURE, target.getUnitTypeKeywords(), this.rangeType, rollGoal);
        }

        if (hasOnesReroll) {
            return unitAttackModifiers.getRerollFor(stageType, RerollType.ON_ONES, target.getUnitTypeKeywords(), this.rangeType, rollGoal);
        }

        return Optional.empty();
    }

    private int calculateWeaponAttacks(int amountOfWeapons, UnitAttackModifiers unitAttackModifiers,
                                          DefensiveProfile target, WeaponAttackResolution weaponAttackResolution) {
        var weaponAttacks = calculateBasicWeaponAttacks(amountOfWeapons, unitAttackModifiers, weaponAttackResolution);

        // Rapid fire
        var hasRapidFire = this.getRapidFire().isPresent() || unitAttackModifiers.getRapidFireModifier()._1;
        if(hasRapidFire && unitAttackModifiers.isTargetInRapidFireRange()) {
            var higherRapidFireValue = Math.max(this.rapidFire.get(), unitAttackModifiers.getRapidFireModifier()._3);
            weaponAttacks += higherRapidFireValue*amountOfWeapons;
            weaponAttackResolution.setAttacksFromRapidFire(higherRapidFireValue);
        }

        // Blast
        if(this.hasBlast) {
            var blastBonus = (int)Math.floor((double)target.getBodies() / 5);
            weaponAttacks += blastBonus*amountOfWeapons;
            weaponAttackResolution.setAttacksFromBlast(blastBonus);
        }

        weaponAttackResolution.setTotalAttacks(weaponAttacks);

        return weaponAttacks;
    }

    private int calculateBasicWeaponAttacks(int amountOfWeapons, UnitAttackModifiers unitAttackModifiers,
                                               WeaponAttackResolution weaponAttackResolution) {
        var weaponAttacks = this.attacks;

        // Attacks modifiers
        var attackModifier = 0;
        var attackBonus = unitAttackModifiers.getAttacksBonusModifier();
        if(attackBonus._1 && attackBonus._2.appliesTo(this)) {
            attackModifier += attackBonus._3;
        }
        var attackMalus = unitAttackModifiers.getAttacksMalusModifier();
        if(attackMalus._1 && attackMalus._2.appliesTo(this)) {
            attackModifier -= attackMalus._3;
        }
        weaponAttackResolution.setAttacksFromModifiers(attackModifier*amountOfWeapons);
        weaponAttacks += attackModifier;

        weaponAttacks *= amountOfWeapons;

        return weaponAttacks;
    }

    private int calculateHitRollGoal(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers) {
        var finalHitRollGoal = this.getHitValue();
        var hitModifier = 0;

        // Bonuses
        var unitHitBonus = unitAttackModifiers.getHitBonusModifier();
        if(unitHitBonus._1 && (unitHitBonus._2.appliesTo(this))) {
            hitModifier--;
        }
        if(this.hasHeavy && unitAttackModifiers.isUnitHasRemainedStationaryModifier()) {
            hitModifier--;
        }

        // Maluses
        var unitHitMalus = unitAttackModifiers.getHitMalusModifier();
        if(unitHitMalus._1 && !unitAttackModifiers.isUnmodifiableHitRoll() && unitHitMalus._2.appliesTo(this)) {
            hitModifier++;
        }
        if(target.isHasStealth() && target.isInStealthRange() && !unitAttackModifiers.isUnmodifiableHitRoll()) {
            hitModifier++;
        }
        if(!target.isVisible() && this.hasIndirectFire && !unitAttackModifiers.isUnmodifiableHitRoll()) {
            hitModifier++;
        }

        // Clever way to limit values to -1,0 or 1, the maximum hit roll modification
        hitModifier = Integer.compare(hitModifier,0);
        finalHitRollGoal += hitModifier;

        return finalHitRollGoal;
    }

    /**
     * Converts from a roll goal to a probability denominator of a fraction P/6. For hit and wound rolls, the die used
     * is D6.
     */
    private static int calculateProbabilityNumeratorFrom(int rollGoal) {
        // Because a roll of 1 always fails and a roll of 6 always succeeds, we modify the roll goal to be between
        // values 2-5
        var modifiedRollGoal = Math.max(rollGoal, 2);
        modifiedRollGoal = Math.min(modifiedRollGoal, 5);

        return 7 - modifiedRollGoal;

    }

    private int calculateWoundRollGoal(DefensiveProfile target, UnitAttackModifiers unitAttackModifiers) {
        var woundRollGoal = 4;
        var finalWoundPenalty = unitAttackModifiers.getWoundMalusModifier()._1 ? 1 : 0;
        var toughness = target.getToughness();

        if(this.strength > toughness) {
            finalWoundPenalty += target.isHasWoundMalusIfStrengthHigherThanToughness() ? 1 : 0;
            woundRollGoal = strength >= toughness*2 ? 2 : 3;
        } else if(strength < toughness) {
            woundRollGoal = toughness*2 >= strength ? 6 : 5;
        }

        var finalWoundBonus = unitAttackModifiers.getWoundBonusModifier()._1 ? 1 : 0;
        var hasLance = (this.hasLance || unitAttackModifiers.isLanceModifier()) && rangeType == WeaponRangeType.MELEE;
        if(unitAttackModifiers.isUnitCharged() &&  hasLance) {
            finalWoundBonus++;
        }

        // Clever way to set either -1, 0 or 1, as the wound roll can only be modified, in the end by 1
        var finalWoundRollModifier = Integer.compare(finalWoundBonus - finalWoundPenalty, 0);

        // Having a -1 to wound rolls implies that the roll goal moves by 1
        woundRollGoal -= finalWoundRollModifier;

        // Now apply any anti-X if it is better than the current wound roll
        var antiWoundRoll = antiUnitEffects.stream()
                .filter(effect -> target.getUnitTypeKeywords().contains(effect.getUnitType()))
                .mapToInt(AntiUnitEffect::getWoundRollValue)
                .min()
                .stream().boxed().findFirst(); // Necessary ugliness
        if(antiWoundRoll.isPresent()) {
            woundRollGoal = Math.min(woundRollGoal, antiWoundRoll.get());
        }

        return woundRollGoal;
    }
}
